@@ -18,7 +18,7 @@ call_block = function(block) {
 
   # expand parameters defined via template
   if (!is.null(block$params$opts.label)) {
-    block$params <- merge_list(opts_template$get(block$params$opts.label), block$params)
+    block$params = merge_list(opts_template$get(block$params$opts.label), block$params)
   }
 
   params = opts_chunk$merge(block$params)
@@ -51,7 +51,7 @@ call_block = function(block) {
       warning("The option hook '", opt, "' should be a function")
       next
     }
-    if (!is.null(params[[opt]])) params = hook(params)
+    if (!is.null(params[[opt]])) params = as.strict_list(hook(params))
     if (!is.list(params))
       stop("The option hook '", opt, "' should return a list of chunk options")
   }
@@ -63,11 +63,13 @@ call_block = function(block) {
       getOption('width'), if (params$cache == 2) params[cache2.opts]
     )
     if (params$engine == 'R' && isFALSE(params$cache.comments)) {
-      content[['code']] = formatR:::parse_only(content[['code']])
+      content[['code']] = parse_only(content[['code']])
     }
     hash = paste(valid_path(params$cache.path, label), digest::digest(content), sep = '_')
     params$hash = hash
-    if (cache$exists(hash, params$cache.lazy) && isFALSE(params$cache.rebuild)) {
+    if (cache$exists(hash, params$cache.lazy) &&
+        isFALSE(params$cache.rebuild) &&
+        params$engine != 'Rcpp') {
       if (opts_knit$get('verbose')) message('  loading cache from ', hash)
       cache$load(hash, lazy = params$cache.lazy)
       if (!params$include) return('')
@@ -76,8 +78,7 @@ call_block = function(block) {
     if (params$engine == 'R')
       cache$library(params$cache.path, save = FALSE) # load packages
   } else if (label %in% names(dep_list$get()) && !isFALSE(opts_knit$get('warn.uncached.dep')))
-    warning('code chunks must not depend on the uncached chunk "', label, '"',
-            call. = FALSE)
+    warning2('code chunks must not depend on the uncached chunk "', label, '"')
 
   params$params.src = block$params.src
   opts_current$restore(params)  # save current options
@@ -102,10 +103,10 @@ block_exec = function(options) {
   if (options$engine != 'R') {
     res.before = run_hooks(before = TRUE, options)
     engine = get_engine(options$engine)
-    output = in_dir(opts_knit$get('root.dir') %n% input_dir(), engine(options))
+    output = in_dir(input_dir(), engine(options))
     res.after = run_hooks(before = FALSE, options)
     output = paste(c(res.before, output, res.after), collapse = '')
-    output = if (is_blank(output)) '' else knit_hooks$get('chunk')(output, options)
+    output = knit_hooks$get('chunk')(output, options)
     if (options$cache) block_cache(
       options, output,
       if (options$engine == 'stan') options$engine.opts$x else character(0)
@@ -118,12 +119,18 @@ block_exec = function(options) {
   obj.before = ls(globalenv(), all.names = TRUE)  # global objects before chunk
 
   keep = options$fig.keep
+  keep.idx = NULL
+  if (is.numeric(keep)) {
+    keep.idx = keep
+    keep = "index"
+  }
+  tmp.fig = tempfile(); on.exit(unlink(tmp.fig), add = TRUE)
   # open a device to record plots
   if (chunk_device(options$fig.width[1L], options$fig.height[1L], keep != 'none',
-                   options$dev, options$dev.args, options$dpi, options)) {
+                   options$dev, options$dev.args, options$dpi, options, tmp.fig)) {
     # preserve par() settings from the last code chunk
     if (keep.pars <- opts_knit$get('global.par'))
-      par(opts_knit$get('global.pars'))
+      par2(opts_knit$get('global.pars'))
     showtext(options$fig.showtext)  # showtext support
     dv = dev.cur()
     on.exit({
@@ -156,20 +163,16 @@ block_exec = function(options) {
   if (keep != 'none' && is.null(options$fig.ext))
     options$fig.ext = dev2ext(options$dev)
 
-  if (!is.null(err.code <- opts_knit$get('stop_on_error'))) {
-    warning('the package option stop_on_error was deprecated;',
-            ' use the chunk option error = ', err.code != 2L, ' instead')
-    options$error = err.code != 2L
-  }
   cache.exists = cache$exists(options$hash, options$cache.lazy)
+  evaluate = knit_hooks$get('evaluate')
   # return code with class 'source' if not eval chunks
   res = if (is_blank(code)) list() else if (isFALSE(ev)) {
     as.source(code)
   } else if (cache.exists && isFALSE(options$cache.rebuild)) {
     fix_evaluate(cache$output(options$hash, 'list'), options$cache == 1)
   } else in_dir(
-    opts_knit$get('root.dir') %n% input_dir(),
-    evaluate::evaluate(
+    input_dir(),
+    evaluate(
       code, envir = env, new_device = FALSE,
       keep_warning = !isFALSE(options$warning),
       keep_message = !isFALSE(options$message),
@@ -201,7 +204,7 @@ block_exec = function(options) {
   if (options$results == 'hide') res = Filter(Negate(is.character), res)
   if (options$results == 'hold') {
     i = vapply(res, is.character, logical(1))
-    if (any(i)) res = c(res[!i], list(paste(unlist(res[i]), collapse = '')))
+    if (any(i)) res = c(res[!i], merge_character(res[i]))
   }
   res = filter_evaluate(res, options$warning, evaluate::is.warning)
   res = filter_evaluate(res, options$message, evaluate::is.message)
@@ -218,6 +221,8 @@ block_exec = function(options) {
         if (keep %in% c('first', 'last')) {
           res = res[-(if (keep == 'last') head else tail)(which(figs), -1L)]
         } else {
+          # keep only selected
+          if (keep == 'index') res = res[which(figs)[keep.idx]]
           # merge low-level plotting changes
           if (keep == 'high') res = merge_low_plot(res, figs)
         }
@@ -233,19 +238,21 @@ block_exec = function(options) {
 
   if (isTRUE(options$fig.beforecode)) res = fig_before_code(res)
 
-  on.exit(plot_counter(reset = TRUE), add = TRUE)  # restore plot number
-  if (options$fig.show != 'animate' && options$fig.num > 1) {
-    options = recycle_plot_opts(options)
-  }
+  on.exit({
+    plot_counter(reset = TRUE)
+    shot_counter(reset = TRUE)
+  }, add = TRUE)  # restore plot number
 
   output = unlist(wrap(res, options)) # wrap all results together
   res.after = run_hooks(before = FALSE, options, env) # run 'after' hooks
 
   output = paste(c(res.before, output, res.after), collapse = '')  # insert hook results
-  output = if (is_blank(output)) '' else knit_hooks$get('chunk')(output, options)
+  output = knit_hooks$get('chunk')(output, options)
 
   if (options$cache > 0) {
-    obj.new = setdiff(ls(globalenv(), all.names = TRUE), obj.before)
+    # if cache.vars has been specifically provided, only cache these vars and no
+    # need to look for objects in globalenv()
+    obj.new = if (is.null(options$cache.vars)) setdiff(ls(globalenv(), all.names = TRUE), obj.before)
     copy_env(globalenv(), env, obj.new)
     objs = if (isFALSE(ev) || length(code) == 0) character(0) else
       options$cache.vars %n% codetools::findLocalsList(parse_only(code))
@@ -256,16 +263,16 @@ block_exec = function(options) {
       dep_auto()
     }
     if (options$cache < 3) {
-      if (!cache.exists) block_cache(options, res.orig, objs)
+      if (options$cache.rebuild || !cache.exists) block_cache(options, res.orig, objs)
     } else block_cache(options, output, objs)
   }
 
-  if (options$include) output else ''
+  if (options$include) output else if (is.null(s <- options$indent)) '' else s
 }
 
 block_cache = function(options, output, objects) {
   hash = options$hash
-  outname = sprintf('.%s', hash)
+  outname = cache_output_name(hash)
   assign(outname, output, envir = knit_global())
   purge_cache(options)
   cache$library(options$cache.path, save = TRUE)
@@ -281,21 +288,29 @@ purge_cache = function(options) {
 
 # open a device for a chunk; depending on the option global.device, may or may
 # not need to close the device on exit
-chunk_device = function(width, height, record = TRUE, dev, dev.args, dpi, options) {
+chunk_device = function(
+  width, height, record = TRUE, dev, dev.args, dpi, options, tmp = tempfile()
+) {
   dev_new = function() {
-    # actually I should adjust the recording device according to dev, but here
-    # I have only considered the png and tikz devices (because the measurement
-    # results can be very different especially with the latter, see #1066)
+    # actually I should adjust the recording device according to dev, but here I
+    # have only considered the png and tikz devices (because the measurement
+    # results can be very different especially with the latter, see #1066), and
+    # also the cairo_pdf device (#1235)
     if (identical(dev, 'png')) {
       do.call(grDevices::png, c(list(
-        filename = tempfile(), width = width, height = height, units = 'in', res = dpi
+        filename = tmp, width = width, height = height, units = 'in', res = dpi
       ), get_dargs(dev.args, 'png')))
     } else if (identical(dev, 'tikz')) {
-      do.call(tikz_dev, c(list(
-        file = paste0(tempfile(), ".tex"), width = width, height = height
-      ), get_dargs(dev.args, 'tikz'), list(
-        sanitize = options$sanitize, standAlone = options$external
-      )))
+      dargs = c(list(
+        file = tmp, width = width, height = height
+      ), get_dargs(dev.args, 'tikz'))
+      dargs$sanitize = options$sanitize; dargs$standAlone = options$external
+      if (is.null(dargs$verbose)) dargs$verbose = FALSE
+      do.call(tikz_dev, dargs)
+    } else if (identical(dev, 'cairo_pdf')) {
+      do.call(grDevices::cairo_pdf, c(list(
+        filename = tmp, width = width, height = height
+      ), get_dargs(dev.args, 'cairo_pdf')))
     } else if (identical(getOption('device'), pdf_null)) {
       if (!is.null(dev.args)) {
         dev.args = get_dargs(dev.args, 'pdf')
@@ -374,9 +389,27 @@ merge_class = function(res, class = c('source', 'message', 'warning')) {
 
 }
 
+# merge character output for output='hold', if the subsequent character is of
+# the same class(es) as the previous one (e.g. should not merge normal
+# characters with asis_output())
+merge_character = function(res) {
+  if ((n <- length(res)) <= 1) return(res)
+  k = NULL
+  for (i in 1:(n - 1)) {
+    cls = class(res[[i]])
+    if (identical(cls, class(res[[i + 1]]))) {
+      res[[i + 1]] = paste0(res[[i]], res[[i + 1]])
+      class(res[[i + 1]]) = cls
+      k = c(k, i)
+    }
+  }
+  if (length(k)) res = res[-k]
+  res
+}
+
 call_inline = function(block) {
   if (opts_knit$get('progress')) print(block)
-  in_dir(opts_knit$get('root.dir') %n% input_dir(), inline_exec(block))
+  in_dir(input_dir(), inline_exec(block))
 }
 
 inline_exec = function(block, envir = knit_global(), hook = knit_hooks$get('inline')) {
@@ -389,7 +422,7 @@ inline_exec = function(block, envir = knit_global(), hook = knit_hooks$get('inli
   for (i in 1:n) {
     v = withVisible(eval(parse_only(code[i]), envir = envir))
     res = if (v$visible) knit_print(v$value, inline = TRUE, options = opts_chunk$get())
-    if (inherits(res, c('knit_asis', 'knit_asis_list'))) res = wrap(res, inline = TRUE)
+    if (inherits(res, 'knit_asis')) res = wrap(res, inline = TRUE)
     d = nchar(input)
     # replace with evaluated results
     stringr::str_sub(input, loc[i, 1], loc[i, 2]) = if (length(res)) {
@@ -443,6 +476,6 @@ label_code = function(code, label) {
          '----', code)
 }
 
-as.source <- function(code) {
+as.source = function(code) {
   list(structure(list(src = code), class = 'source'))
 }
